from opencv.cv import *
from opencv.highgui import *


import serial
import time
import msvcrt
import math
import sys

DEBUG = 0
NOSEND = 0
RIGHT = -2      #This number is arbitrary, but the same as used in oo_remote.py

UPDATE                  = 0x1A
SET_MOTOR_COMMAND       = 0x77
RESET                   = 0x39
SET_LIMIT_SWITCH_MODE   = 0x80
SET_PROFILE_MODE        = 0xA0
SET_ACCELERATION        = 0x90
SET_DECELERATION        = 0x91
SET_JERK                = 0x13
SET_VELOCITY            = 0x11
GET_COMMANDED_POSITION  = 0x1D  #Get Instantaneous Position
SET_ACTUAL_POSITION_UNITS=0xBE

LastNumber0 = 0
LastNumber1 = 0

TURNING = 3.
MOVING = 0.5


def hibyte(word):
    return word/256

def lobyte(word):
    return word - (hibyte(word) * 256)

def loword(dword):
    return dword - (hiword(dword) * 65536)

def hiword(dword):
    return dword/65536

def ordIF(s1, s2):
    c = (ord(s1) - ord(s2))
    if c > 9:
        c -= 7
    return c

def toByte(s):
    s = s.upper()
    return (ordIF(s[0], "0")*16) + ordIF(s[1], "0")

class PWDCmd:
    address = 0x0
    checksum = 0x0
    axis = 0        #Always zero
    code = 0x0
    data = [0xCC,0xCC,0xCC,0xCC,0xCC,0xCC]

class Pos:
    x = 0.
    y = 0.
    theta = 0.      # FROM ZERO TO 2PI, std. unit circle
    CIR_DIST = 10.16*math.pi*4.1 #About 4.1 turns of the wheel, diameter of wheel ~10cm

    #Assumes that diff0 ~= diff1, which is not actually true
    def Adjust(self, diff0, diff1):
        if diff1 == 0 and diff0 == 0:
            pass
        elif (diff1 > 0 and diff0 > 0) or (diff1 < 0 and diff0 < 0):   #forward/backward
            self.x += math.cos(self.theta)*diff0
            self.y += math.sin(self.theta)*diff0
        else:       #turning
            self.theta += (diff0/self.CIR_DIST)*math.pi*2
            while self.theta > math.pi*2.:
                self.theta -= math.pi*2.
            while self.theta < 0:
                self.theta += math.pi*2.

pos = Pos()
ser = serial.Serial()

def SetCmd2(byte1, byte2):
    cmd = PWDCmd()
    cmd.data = [0xCC,0xCC,0xCC,0xCC,0xCC,0xCC]  #for some reason this gets changed in the class, so always reset to default
    cmd.address = byte1
    cmd.code = byte2
    cmd.checksum = ~(byte1 + byte2) + 1
    if cmd.checksum < 0:
        cmd.checksum += 256
    return cmd

def SetCmd3(byte1, byte2, data1):
    cmd = SetCmd2(byte1, byte2)
    cmd.data[0] = hibyte(data1)
    cmd.data[1] = lobyte(data1)
    cmd.checksum = ~(~(cmd.checksum-1)+ cmd.data[0] + cmd.data[1]) + 1
    if cmd.checksum < 0:
        cmd.checksum += 256
    return cmd;

def SetCmd4(byte1, byte2, data1, data2):
    cmd = SetCmd3(byte1, byte2, data1)
    cmd.data[2] = hibyte(data2)
    cmd.data[3] = lobyte(data2)
    cmd.checksum = ~(~(cmd.checksum-1)+ cmd.data[2] + cmd.data[3]) + 1
    if cmd.checksum < 0:
        cmd.checksum += 256
    return cmd

def SendCmd(cmd):
    reply = ""
    if not NOSEND:
        bytes = ""
        bytes += chr(cmd.address)
        bytes += chr(cmd.checksum)
        bytes += chr(cmd.axis)
        bytes += chr(cmd.code)
        for b in cmd.data:
            if b < 0:
                b += 256
            if b != 0xCC:       # For whatever reason 0xCC is "null" -> we could fix this by keeping track of size...
                bytes += chr(b)
        ser.write(bytes)
        time.sleep(0.03)        # Wait for response to be created
        reply = ser.read(size=8)
            
    if DEBUG:
        bytes = ""
        bytes += str(cmd.address)
        bytes += " "
        bytes += str(cmd.checksum)
        bytes += " "
        bytes += str(cmd.axis)
        bytes += " "
        bytes += str(cmd.code)
        bytes += " "
        for b in cmd.data:
            if b < 0:
                b += 256
            bytes += str(b) + " "
        print bytes
        if not NOSEND:
            newstring = "Reply "
            for s in reply:
                newstring += str(ord(s)) + " "
            print newstring
    return reply

def SendCmd2(byte1, byte2):
    return SendCmd(SetCmd2(byte1, byte2))

def SendCmd3(byte1, byte2, data1):
    return SendCmd(SetCmd3(byte1, byte2, data1))

def SendCmd4(byte1, byte2, data1, data2):
    return SendCmd(SetCmd4(byte1, byte2, data1, data2))

def WheelUpdate():
    SendCmd2(1,UPDATE)
    SendCmd2(0,UPDATE)

def WheelSetPower(data):
    SendCmd3(1, SET_MOTOR_COMMAND, data)
    SendCmd3(0, SET_MOTOR_COMMAND, data)

def WheelInit():
    SendCmd2(1, RESET)
    SendCmd3(1, SET_LIMIT_SWITCH_MODE, 0x0000)
    SendCmd3(1, SET_MOTOR_COMMAND, 0x0000)
    SendCmd3(1, SET_PROFILE_MODE, 0x0001)
    SendCmd4(1, SET_ACCELERATION, 0x0000, 0x00B0)
    SendCmd4(1, SET_DECELERATION, 0x0000, 0x00B0)
    SendCmd4(1, SET_JERK, 0x0329, 0x0ABB)

    SendCmd2(0, RESET)
    SendCmd3(0, SET_LIMIT_SWITCH_MODE, 0x0000)
    SendCmd3(0, SET_MOTOR_COMMAND, 0x0000)
    SendCmd3(0, SET_PROFILE_MODE, 0x0001)
    SendCmd4(0, SET_ACCELERATION, 0x0000, 0x00B0)
    SendCmd4(0, SET_DECELERATION, 0x0000, 0x00B0)
    SendCmd4(0, SET_JERK, 0x0329, 0x0ABB)

    WheelUpdate()

def WheelSetVelocity(dwSpeed):
    dwNegSpeed = ~dwSpeed + 1
    WheelSetPower(0x4CC0)
    SendCmd4(1, SET_VELOCITY, hiword(dwNegSpeed), loword(dwNegSpeed))
    SendCmd4(0, SET_VELOCITY, hiword(dwSpeed), loword(dwSpeed))

def WheelSetTurn(dwSpeed):    # + is left, neg is right
    WheelSetPower(0x4CC0)
    SendCmd4(1, SET_VELOCITY, hiword(dwSpeed), loword(dwSpeed))
    SendCmd4(0, SET_VELOCITY, hiword(dwSpeed), loword(dwSpeed))

def GetPos():
    # In theory, these results are in counts, not that we know (yet) what they mean!
    # But the numbers are so large, they must be in microsteps...
    reply = SendCmd2(1, GET_COMMANDED_POSITION)
    number1 = 0
    i = 3
    for s in reply[2:5]:
        number1 += ord(s)*(256**i)
        i -= 1
    number1 = -number1
    number1 += 0xFFFFFFFF
    if number1 > (0xFFFFFFFF/2): #Due to the limit of the PMD processor, we only have a ulong. So, half is pos, half is neg.
	# That is to say, we are using two's complement arithmetic on a 16-bit processor. 
        number1 -= 0xFFFFFFFF
    number1 /= 64.       # Number of microsteps per electrical cycle
    number1 /= 256.      # Number of steps?
    
    reply = SendCmd2(0, GET_COMMANDED_POSITION)
    number0 = 0
    i = 3
    for s in reply[2:5]:
        number0 += ord(s)*(256**i)
        i -= 1
    if number0 > (0xFFFFFFFF/2):
        number0 -= 0xFFFFFFFF
    number0 /= 64.       # Number of microsteps per electrical cycle
    number0 /= 256.
    
    global LastNumber0, LastNumber1
    diff1 = number1 - LastNumber1
    diff0 = number0 - LastNumber0
    diff1 *= 10.16*math.pi/3.28
    diff0 *= 10.16*math.pi/3.28

    LastNumber1 = number1
    LastNumber0 = number0
    pos.Adjust(diff0, diff1)

def PrintPos():
    GetPos()
    string = "Position: " + str(pos.x) + " " + str(pos.y) + " " + str(pos.theta)
    print string

def InitializeRobot(PORT):
    ser.port = PORT-1
    ser.baudrate = 250000
    ser.timeout = 0
    ser.open()
    if not ser.isOpen():
        print "ERROR OPENING ROBOT PORT"
        sys.exit(-1)

    WheelInit()

def Go(speed):   #Speed is cm/s
    WheelSetVelocity(int(speed*16700))
    WheelUpdate()
    GetPos()

def GoDist(dist, speed=10.):   #dist in cm
    if dist < 0 and speed > 0:
        speed = -speed
    startX = curX = pos.x
    startY = curY = pos.y
    while abs(dist) - math.sqrt((startX-curX)**2 + (startY-curY)**2) > abs(speed)/2.5:
        Go(speed)
        GetPos()
        curX = pos.x
        curY = pos.y
    Go(0)

def Turn(speed, direction):
    if direction == RIGHT:
        speed = -speed
    WheelSetTurn(int(speed*16700))
    WheelUpdate()
    GetPos()

def TurnAngle(radians, direction, speed=8.):
    if direction == RIGHT:
        radians = -radians
    wantTh = radians + pos.theta
    while wantTh > math.pi*2.:
        wantTh -= math.pi*2.
    while wantTh < 0.:
        wantTh += math.pi*2.
    curTh = pos.theta
    while abs(wantTh - curTh) > (math.pi*2./pos.CIR_DIST) * speed/2.5:
        Turn(speed, direction)
        GetPos()
        curTh = pos.theta
    Turn(0, direction)

def GetRobotPos():
    GetPos()
    string = str(pos.x) + " " + str(pos.y) + " " + str(pos.theta)
    return string
    
def CloseRobot():
    WheelSetVelocity(0x0)
    WheelUpdate()
    WheelSetPower(0x0)
    ser.close()
