//
// The full "Square Detector" program.
// It loads several images subsequentally and tries to find squares in
// each image
//

#include "stdafx.h"

#include "squares.h"
#include <stdio.h>
#include <math.h>
#include <string>
#include <fstream>
#include <highgui.h>
#include <iostream>

using namespace std;

int thresh = 50;
int count = 0;
int imgNum = 0;

// helper function:
// finds a cosine of angle between vectors
// from pt0->pt1 and from pt0->pt2 
double angle( CvPoint* pt1, CvPoint* pt2, CvPoint* pt0 )
{
    double dx1 = pt1->x - pt0->x;
    double dy1 = pt1->y - pt0->y;
    double dx2 = pt2->x - pt0->x;
    double dy2 = pt2->y - pt0->y;
    return (dx1*dx2 + dy1*dy2)/sqrt((dx1*dx1 + dy1*dy1)*(dx2*dx2 + dy2*dy2) + 1e-10);
}

// returns sequence of squares detected on the image.
// the sequence is stored in the specified memory storage
CvSeq* findSquares4( IplImage* img, CvMemStorage* storage )
{
    CvSeq* contours;
    int i, c, l, N = 11;
    CvSize sz = cvSize( img->width & -2, img->height & -2 );
    IplImage* timg = cvCloneImage( img ); // make a copy of input image
    IplImage* gray = cvCreateImage( sz, 8, 1 ); 
    IplImage* pyr = cvCreateImage( cvSize(sz.width/2, sz.height/2), 8, 3 );
    IplImage* tgray;
    CvSeq* result;
    double s, t;
    // create empty sequence that will contain points -
    // 4 points per square (the square's vertices)
    CvSeq* squares = cvCreateSeq( 0, sizeof(CvSeq), sizeof(CvPoint), storage );
    
    // select the maximum ROI in the image
    // with the width and height divisible by 2
    cvSetImageROI( timg, cvRect( 0, 0, sz.width, sz.height ));
    
    // down-scale and upscale the image to filter out the noise
    cvPyrDown( timg, pyr, 7 );
    cvPyrUp( pyr, timg, 7 );
    tgray = cvCreateImage( sz, 8, 1 );
    
    // find squares in every color plane of the image
    for( c = 0; c < 1; c++ )
    {
        // extract the c-th color plane
        cvSetImageCOI( timg, c+1 );
        cvCopy( timg, tgray, 0 );
        
        // try several threshold levels
        for( l = 0; l < N; l++ )
        {
            // hack: use Canny instead of zero threshold level.
            // Canny helps to catch squares with gradient shading   
            if( l == 0 )
            {
                // apply Canny. Take the upper threshold from slider
                // and set the lower to 0 (which forces edges merging) 
                cvCanny( tgray, gray, 50, 150, 5 );
                // dilate canny output to remove potential
                // holes between edge segments
                cvDilate( gray, gray, 0, 1 );
            }
            else
            {
                // apply threshold if l!=0:
                //     tgray(x,y) = gray(x,y) < (l+1)*255/N ? 255 : 0
                cvThreshold( tgray, gray, (l+1)*255/N, 255, CV_THRESH_BINARY );
            }
            
            // find contours and store them all as a list
            cvFindContours( gray, storage, &contours, sizeof(CvContour),
                CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, cvPoint(0,0) );
            
            // test each contour
            while( contours )
            {
                // approximate contour with accuracy proportional
                // to the contour perimeter
                result = cvApproxPoly( contours, sizeof(CvContour), storage,
                    CV_POLY_APPROX_DP, cvContourPerimeter(contours)*0.05, 0 );
                // square contours should have 4 vertices after approximation
                // relatively large area (to filter out noisy contours)
                // and be convex.
                // Note: absolute value of an area is used because
                // area may be positive or negative - in accordance with the
                // contour orientation
#define BORDER 10
                if( result->total == 4 &&
					fabs(cvContourArea(result,CV_WHOLE_SEQ)) > sz.height*sz.width*0.01 &&
					fabs(cvContourArea(result,CV_WHOLE_SEQ)) < sz.height*sz.width*0.85 &&
					((CvPoint*)cvGetSeqElem( result, 0 ))->x > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->x < img->width-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 1 ))->x > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->x < img->width-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 2 ))->x > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->x < img->width-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 3 ))->x > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->x < img->width-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 0 ))->y > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->y < img->height-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 1 ))->y > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->y < img->height-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 2 ))->y > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->y < img->height-BORDER &&
					((CvPoint*)cvGetSeqElem( result, 3 ))->y > BORDER && ((CvPoint*)cvGetSeqElem( result, 0 ))->y < img->height-BORDER &&
					cvCheckContourConvexity(result) )
				{
#if 0
					s = 0;
                    
                    for( i = 0; i < 5; i++ )
                    {
                        // find minimum angle between joint
                        // edges (maximum of cosine)
                        if( i >= 2 )
                        {
                            t = fabs(angle(
                            (CvPoint*)cvGetSeqElem( result, i ),
                            (CvPoint*)cvGetSeqElem( result, i-2 ),
                            (CvPoint*)cvGetSeqElem( result, i-1 )));
                            s = s > t ? s : t;
                        }
                    }
                    
                    // if cosines of all angles are small
                    // (all angles are ~90 degree) then write quandrange
                    // vertices to resultant sequence 
                    if( s < 0.4 )
#endif
                        for( i = 0; i < 4; i++ )
							cvSeqPush( squares, (CvPoint*)cvGetSeqElem( result, i ));
				}
				
				// take the next contour
				contours = contours->h_next;
            }
        }
    }
    
    // release all the temporary images
    cvReleaseImage( &gray );
    cvReleaseImage( &pyr );
    cvReleaseImage( &tgray );
    cvReleaseImage( &timg );

    return squares;
}


// the function draws all the squares in the image
IplImage* drawSquares( IplImage* img, CvSeq* squares, const char* filename )
{
    CvSeqReader reader;
	IplImage* cpy = cvCloneImage( img );
    int i;
    
    // initialize reader of the sequence
    cvStartReadSeq( squares, &reader, 0 );
    
    CvPoint maxRect[4];
	double maxRectArea = 0.0;
	double area;
	
	// read 4 sequence elements at a time (all vertices of a square)
    for( i = 0; i < squares->total; i += 4 )
    {
        CvPoint pt[4], *rect = pt;
        int count = 4;
        
        // read 4 vertices
        CV_READ_SEQ_ELEM( pt[0], reader );
        CV_READ_SEQ_ELEM( pt[1], reader );
        CV_READ_SEQ_ELEM( pt[2], reader );
        CV_READ_SEQ_ELEM( pt[3], reader );
        
        // find the area of the quadrilateral
		area = fabs(1.0*(pt[1].x*pt[0].y - pt[0].x*pt[1].y)+ (pt[2].x*pt[1].y - pt[1].x*pt[2].y)+(pt[0].x*pt[2].y - pt[2].x*pt[0].y))/2
			 + fabs(1.0*(pt[1].x*pt[3].y - pt[3].x*pt[1].y)+ (pt[2].x*pt[1].y - pt[1].x*pt[2].y)+(pt[3].x*pt[2].y - pt[2].x*pt[3].y))/2;
		
		if(area > maxRectArea)
		{
			maxRectArea = area;
			maxRect[0] = pt[0];
			maxRect[1] = pt[1];
			maxRect[2] = pt[2];
			maxRect[3] = pt[3];
		}
		
		// draw the square as a closed polyline 
        cvPolyLine( cpy, &rect, &count, 1, 1, CV_RGB(0,255,0), 3, CV_AA, 0 );
    }
    
	ofstream out(filename);
	if(maxRectArea > 0.0)
	{
		out << (maxRect[0].x+maxRect[1].x+maxRect[2].x+maxRect[3].x)/(2.0*img->width) - 1.0 << endl;
		out << (maxRect[0].y+maxRect[1].y+maxRect[2].y+maxRect[3].y)/(2.0*img->height) - 1.0 << endl;		
	}
	else
		out << " ";
	out.close();

	if (PRINT_CAM_PROCESSED)
	{
//		if (count == 10)
//		{
			char tempBuffer[50];
			itoa(imgNum, tempBuffer, 10);
			string tempBuffer2 = tempBuffer;
			string tempBuffer3(filename);
			tempBuffer3.replace(0, 2, "c:\\images\\");

			string debugName = tempBuffer3;
			debugName += "processed";
			debugName += tempBuffer2;
			debugName += ".jpg";

			count = 0;

			imgNum++;

			cvSaveImage(debugName.c_str(), (CvArr*)cpy);
//		}
//		else
//			count++;
	}


	// show the resultant image
    return cpy;
}